/***************************************************************************/
/*                                                                         */
/*                           Repalette handler                             */
/*                                                                         */
/*                                                                         */
/***************************************************************************/

/* CLib */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Application modules */

#include "kisshdr.h"

#define REPALETTE_K_OUTOFMEM    1
#define REPALETTE_K_MISMPALETTE 2
#define REPALETTE_K_SAVECEL     3
#define REPALETTE_K_NOTCHERRY   4
#define REPALETTE_K_READCEL     5
#define REPALETTE_K_READKCF     6
#define REPALETTE_K_READKCF2    7

/**************************************************************************/
/*
 * Compute the colour distance from one palette entry to another. Uses
 * a simple weighted sum og squares for each component.
 *
 * (Algorithm/weights as in RISCOS colourtrans module)
 */

#define RED_WEIGHT   2
#define GREEN_WEIGHT 4
#define BLUE_WEIGHT  1

static int colourdistance(paletteword palette,paletteword altpal)
{
  int r,g,b;

  r = palette.bytes.red   - altpal.bytes.red;
  g = palette.bytes.green - altpal.bytes.green;
  b = palette.bytes.blue  - altpal.bytes.blue;

  return (r * r * RED_WEIGHT) +
         (g * g * GREEN_WEIGHT) +
         (b * b * BLUE_WEIGHT);
}

/**************************************************************************/
/*
 * Create a pixel colour translation table from one palette to another.
 */

extern byte *repalette_CreatePixelTransTable(paletteword *palette,
                                             int          numold,
                                             paletteword *newpalette,
                                             int          numnew)
{
  int newcol, oldcol, bestcol, besterr, err;
  byte *ptt;
  paletteword *newpal;

  ptt = malloc(numold);
  if ( !ptt )
    return NULL;

  /*
   * Translation uses rules as follows:
   *    Colour 0 always maps to colour zero.
   *    All other colours map to the visually-nearest colour in the
   *    other palette, EXCEPT that no colour maps onto zero.
   */

  ptt[0] = 0;
  palette++;
  newpalette++;

  for ( oldcol = 1; oldcol < numold; oldcol++, palette++ )
  {
    bestcol = 1;
    newpal  = newpalette;
    besterr = colourdistance(*palette,*newpal);
    newpal++;

    for ( newcol = 2; newcol < numnew; newcol++, newpal++ )
    {
      err = colourdistance(*palette,*newpal);
      if ( err < besterr )
      {
        bestcol = newcol;
        besterr = err;
      }
    }
    ptt[oldcol] = bestcol;
  }

  return ptt;
}

/**************************************************************************/
/*
 * Dispose of the malloc'ed memory used to store cel pixels.
 */

extern void repalette_DisposeCel(CEL_DESCRIPTOR *cel)
{
  int y;

  for ( y = 0; y < cel->hdr.height; y++ )
    free(cel->rows[y]);

  free(cel->rows);
}

/**************************************************************************/

static int create_pixel_array(CEL_DESCRIPTOR *cel)
{
  int y;

  switch ( cel->hdr.bpp )
  {
    case 4 : cel->rowsize = (cel->hdr.width+1)>>1;
             break;
    case 8 : cel->rowsize = cel->hdr.width;
             break;
    default : return REPALETTE_K_READCEL;
  }

  if ( cel->rows = malloc(cel->hdr.height * sizeof(byte *)), cel->rows == NULL)
    return REPALETTE_K_OUTOFMEM;

  memset(cel->rows,0,cel->hdr.height * sizeof(byte *));

  for ( y = 0; y < cel->hdr.height; y++ )
    if ( cel->rows[y] = malloc(cel->rowsize), cel->rows[y] == NULL )
      break;

  if ( y != cel->hdr.height )
  {
    repalette_DisposeCel(cel);
    return REPALETTE_K_OUTOFMEM;
  }
  return 0;
}

/**************************************************************************/
/*
 * Perform the actual pixel translation from one cel to another.
 */

extern int repalette_ProcessCel( CEL_DESCRIPTOR *cel,
                                 byte           *pixtrans,
                                 int             newnumcols,
                                 CEL_DESCRIPTOR *newcel)
{
  int x, y, status;
  byte *newrow, *oldrow, newcol, colno;

  /*
   * Create an appropriate header block for the new cel
   */

  if ( (newcel->hdr.bpp==8) && !cel->gs2 )
  {
    newcel->gs2            = 1;
    memset(&(newcel->hdr),0,sizeof(KISS_CEL_HEADER));

    newcel->hdr.kissdataid = KISS_ID;
    newcel->hdr.mark       = KISS_CEL_MARK;
    newcel->hdr.bpp        = 8;
    newcel->hdr.width      = cel->hdr.width;
    newcel->hdr.height     = cel->hdr.height;
    newcel->rows = NULL;
  }
  else memcpy(newcel,cel,sizeof(CEL_DESCRIPTOR));

  newcel->hdr.bpp = (newnumcols == 16) ? 4 : 8;

  if ( status = create_pixel_array(newcel), status != 0 )
    return status;

  /*
   * Now copy the old cel to the new cel, with allowance for the
   * colour conversion provided in the pixel translation table.
   */

  for ( y = 0; y < newcel->hdr.height; y++ )
  {
    newrow = newcel->rows[y];
    oldrow = cel->rows[y];

    for ( x = 0; x < newcel->hdr.width; x++ )
    {
      if (cel->hdr.bpp == 4)
      {
        colno = oldrow[x>>1];
        colno = (x & 1) ? (colno & 0xf) : ((colno>>4) & 0xf);
      }
      else colno = oldrow[x];

      newcol = pixtrans[colno];

      if ( newcel->hdr.bpp == 4 )
      {
        colno = newrow[x>>1];
        if ( x & 1 )
          newcol = (colno & 0xf0) | newcol;
        else newcol = (colno & 0x0f) | (newcol << 4);
        newrow[x>>1] = newcol;
      }
      else newrow[x] = newcol;
    }
  }
  return 0;
}

/**************************************************************************/
/*
 * Save a cel.
 */

extern int repalette_SaveCelFile(char           *filename,
                                 CEL_DESCRIPTOR *cel)
{
  FILE *fp;
  int y, status;

  if ( fp = fopen(filename,"wb"), fp == NULL)
    return REPALETTE_K_SAVECEL;

  if ( cel->gs2 )
    status = fwrite(&cel->hdr,sizeof(KISS_CEL_HEADER),1,fp) == 1;
  else
   status = fwrite(&cel->hdr.width,sizeof(int16)*2,1,fp) == 1;

  if ( !status )
  {
    fclose(fp);
    return REPALETTE_K_SAVECEL;
  }

  for ( y = 0; (y < cel->hdr.height) & status; y++ )
    status = fwrite(cel->rows[y],cel->rowsize,1,fp) == 1;

  fclose(fp);

  return status ? 0 : REPALETTE_K_SAVECEL;
}

/**************************************************************************/
/*
 * Read a palette file into memory.
 */

extern int repalette_ReadPaletteFile(char         *filename,
                                     int          grp,
                                     int          *numcols,
                                     paletteword **palette,
                                     int          reqdcols)
{
  KISS_PAL_HEADER hdr;
  int id, entrysize, shift, mask, palPosition, colno, shift2, status;
  paletteword *pal;
  int32 col;
  FILE *fp;

  *palette = NULL;
  *numcols = 0;

  if ( fp = fopen(filename,"rb"), fp == NULL )
    return REPALETTE_K_READKCF;

  if ( fread(&id,sizeof(int),1,fp) != 1 )
  {
    fclose(fp);
    return REPALETTE_K_READKCF;
  }

  fseek(fp,0l,SEEK_SET);

  if ( id == KISS_ID )
  {
    if ( fread(&hdr,sizeof(KISS_PAL_HEADER),1,fp) != 1 )
    {
      fclose(fp);
      return REPALETTE_K_READKCF;
    }
  }
  else
  {
    memset(&hdr,0,sizeof(KISS_PAL_HEADER));
    hdr.bpcol      = 12;
    hdr.numgroup   = 10;
    hdr.colspergrp = 16;
  }
  if ( grp >= hdr.numgroup )
    grp = hdr.numgroup-1;

  palPosition = hdr.colspergrp * ((hdr.bpcol == 12) ? 2 : 3 );
  palPosition *= grp;

  if ( fseek(fp,palPosition,SEEK_CUR) )
  {
    fclose(fp);
    return REPALETTE_K_READKCF;
  }

  if ( hdr.bpcol == 12 )
  {
    entrysize = 2;
    mask      = 0xf;
    shift     = 4;
    shift2    = 4;
  }
  else
  {
    entrysize = 3;
    mask      = 0xff;
    shift     = 8;
    shift2    = 0;
  }

  if ( pal = malloc(hdr.colspergrp * sizeof(paletteword)), pal == NULL )
  {
    fclose(fp);
    return REPALETTE_K_OUTOFMEM;
  }
  memset(pal,0,hdr.colspergrp * sizeof(paletteword));

  *palette = pal;
  *numcols = hdr.colspergrp;
  if ( (reqdcols>0) && (reqdcols != hdr.colspergrp) )
    status = REPALETTE_K_MISMPALETTE;
  else status = 0;

  for ( colno = 0; colno < hdr.colspergrp; colno++ )
  {
    if ( status = fread(&col,entrysize,1,fp) != 1, status != 0)
      break;

    if ( hdr.bpcol == 12 )
      col =  ((col >> 4) & 0xff) | ((col & 0xf)<<8);

    pal->bytes.red    = (byte)(( col & mask) << shift2);
    pal->bytes.green  = (byte)(( (col >> shift) & mask ) << shift2);
    pal->bytes.blue   = (byte)(( (col>>(shift<<1)) & mask ) << shift2);
    pal->bytes.unused = 0;
    pal++;
  }

  fclose(fp);

  return status;
}

/**************************************************************************/
/*
 * Read a cel into memory.
 */

extern int repalette_ReadCelFile(char           *filename,
                                 CEL_DESCRIPTOR *cel)
{
  int status;
  int32 key;
  FILE *fp;
  int row;

  if ( fp = fopen(filename,"rb"), fp == NULL )
    return REPALETTE_K_READCEL;

  /*
   * Read & decode the KISS header block
   */

  if ( fread(&key,sizeof(int32),1,fp) != 1 )
  {
    fclose(fp);
    return REPALETTE_K_READCEL;
  }

  fseek(fp,0l,SEEK_SET);

  cel->gs2 = 0;

  if ( key == KISS_ID )
  {
    status = fread(&(cel->hdr),sizeof(KISS_CEL_HEADER),1,fp) == 1;
    cel->gs2 = 1;
  }
  else
    {
      status        = fread(&(cel->hdr.width),sizeof(int16),2,fp) == 2;
      cel->hdr.bpp  = 4;
      cel->hdr.mark = KISS_CEL_MARK;
    }

  if ( !status || (cel->hdr.mark != KISS_CEL_MARK) )
  {
    fclose(fp);
    return (cel->hdr.mark == KISS_C32_MARK) ? REPALETTE_K_NOTCHERRY
                                            : REPALETTE_K_READCEL;
  }

  if ( status = create_pixel_array(cel), status != 0 )
  {
    fclose(fp);
    return status;
  }

  for ( row = 0; row < cel->hdr.height; row++ )
    if ( fread(cel->rows[row],cel->rowsize,1,fp) != 1 )
    {
      status = REPALETTE_K_READCEL;
      break;
    }

  fclose(fp);

  if ( status )
    repalette_DisposeCel(cel);

  return status;
}

/**************************************************************************/
/*
 * Highest-level possible interface to repalette functions.
 */

extern int repalette_ProcessFile(char  *celfile,
                                 char  *oldpalfile,
                                 char  *newpalfile,
                                 char  *outfile)
{
  int status, oldsize, newsize;
  CEL_DESCRIPTOR cel,newcel;
  paletteword *oldpal=NULL,*newpal=NULL;
  byte *ptt;

  status = repalette_ReadCelFile(celfile,&cel);
  if ( status == 0 )
  {
    status = repalette_ReadPaletteFile(oldpalfile,
                                       0,
                                       &oldsize,
                                       &oldpal,
                                       1<<cel.hdr.bpp);

    if ( status == 0 )
    {
      status = repalette_ReadPaletteFile(newpalfile,0,&newsize,&newpal,0);
      if ( status == REPALETTE_K_READKCF )
        status = REPALETTE_K_READKCF2;

      if ( status == 0 )
      {
        ptt = repalette_CreatePixelTransTable(oldpal,oldsize,newpal,newsize);

        if ( status == 0 )
        {
          status = repalette_ProcessCel(&cel,ptt,newsize,&newcel);

          if ( status == 0 )
          {
            status = repalette_SaveCelFile(outfile,&newcel);
            repalette_DisposeCel(&newcel);
          }

          free(ptt);
        }

        free(newpal);
      }

      free(oldpal);
    }
    repalette_DisposeCel(&cel);
  }

  return status;
}
